//
//  pvlog.c
//  
//
//  Created by Brian Cornell on 10/9/17.
//
//
//


#define _LCLIENT                        //Required for logger client programs
//#define _BSD                          //Uncomment when compiling for FreeBSD

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <strings.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>

#ifdef _BSD
#include "logger.h"
#else
#include "../logger_1-1.X/logger.h"
#endif

#include "pvlog.h"


//Module variables so signal hander has access:  only functions main & sigHandler should directly access.
FILE *LogFile = NULL;                   //Data log file handle
dats DATfiles;                          //Output data files in log mode
char *prgName = STR_PRGNAME;


int main(int argc, char *argv[])
{
    logft LogFtr;                       //Footer record
    logag LogAng;                       //Analog data record
    time_t LogEpoch_t;                  //Time that logger was initialized
    time_t SysTime_t;                   //Used to fetch system time
    int ReturnCode, Cmd;
    unsigned int RecCount;
    char *lfmode = FMODE_LOG_FIL;

    DATfiles.log = stderr;              //Default is to log to terminal
    while (TRUE)
    {
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                       Argument processing
         * If no arguments display program info & exit.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        ReturnCode = RC_NOARG;
        if (argc < SIZ_ARGC_MIN)
        {
            printf(FMT_PRG_DESC, VERMAJOR, VERMINOR, VERBUILD);
            printf(FMT_PRG_SYNTX, prgName);
            for (Cmd = 0; (Cmd < SIZE_A_CMD); ++Cmd)\
                printf(FMT_PRG_CMDL, Commands[Cmd].str, Commands[Cmd].desc);
            break;
        }

        //Process command-line arguments
        ReturnCode = RC_BADARG;
        if (argc > SIZ_ARGC_MIN)
        {
            for (Cmd = 0; (Cmd < SIZE_A_CMD); ++Cmd)\
                if (strcasecmp(argv[SIZ_ARGV_CMD], Commands[Cmd].str) == 0) break;
            if (Cmd == SIZE_A_CMD) break;           //No valid command found
        }
        else Cmd = CMD_RAW;                         //Default if only input file given

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                         Logging mode
         * Requires path & file name base to output files.  Initialize data 
         * files and set log file open mode to r/w.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (Cmd == CMD_LOG)
        {
            if (argc < (int) SIZ_ARGC_BASE) break;

            ReturnCode = RC_DATCFAIL;
            if (openDAT(&DATfiles, argv[SIZ_ARGV_BASE]) == NULL) break;
            lfmode = FMODE_LOG_DEV;
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                        Finish initialization
         * Open log device or file; exit on fail.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        ReturnCode = RC_LOGOFAIL;
        if ((LogFile = fopen(argv[1], lfmode)) == NULL) break;

        //Install signal handlers
        ReturnCode = RC_SIGERROR;
        if (signal(SIGINT, sigHandler) == SIG_ERR) break;
        if (signal(SIGTERM, sigHandler) == SIG_ERR) break;
        if (signal(SIGHUP, sigHandler) == SIG_ERR) break;
        if (signal(SIGQUIT, sigHandler) == SIG_ERR) break;

        //Start logging if reading from device
        if (Cmd == CMD_LOG)
        {
            ReturnCode = RC_LOGDFAIL;
            if (startLog(LogFile) == FALSE) break;
        }

        //Read header record & validate; exit on fail
        ReturnCode = RC_LOGHFAIL;
        if ((LogEpoch_t = readHeader(LogFile)) == (time_t) -1) break;

        //If not logging print field header record
        if (Cmd == CMD_DUMP) fprintf(stdout, FMT_HDR_FIELDS, STR_HDR_RAW,\
                                     STR_HDR_DMP);
        else if (Cmd == CMD_PROC) fprintf(stdout, FMT_HDR_FIELDS, STR_HDR_PRC,\
                                          STR_HDR_DMP);
        else if (Cmd == CMD_RAW) fprintf(stdout, STR_HDR_RAW);

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                         *** Process records ***
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        for (RecCount = 0, ReturnCode = RC_LOGAFAIL;\
             (fread((void *) &LogAng, sizeof(logag), 1, LogFile)); ++RecCount,\
             ReturnCode = RC_LOGAFAIL)
        {
            ReturnCode = RC_RECTYPE;
            if (LogAng.rectype != RECTANL)
            {
                int c;
                unsigned char *p;
                fprintf(DATfiles.log, FMT_ERR_RECT);
                for (c = 0, p = (unsigned char *) &LogAng; (c < sizeof(logag));\
                     ++c, ++p) fprintf(DATfiles.log, FMT_ERR_RECTV, (unsigned int) *p);
                break;
            }
            if (LogAng.seqnum != RecCount)
            {
                ReturnCode = RC_SEQNUMBR;
                fprintf(DATfiles.log, FMT_ERR_SEQ, RecCount, (unsigned int) LogAng.seqnum);
                break;
            }

            //Validation complete - analyze / process record.
            if (Cmd == CMD_DUMP) DumpRec(stdout, &LogAng, &LogEpoch_t);
            else if (Cmd == CMD_PROC) ProcessRec(stdout, &LogAng, &LogEpoch_t);
            else if (Cmd == CMD_RAW) RawRec(stdout, &LogAng, &LogEpoch_t);
            else if (Cmd == CMD_LOG)
            {
                //First process & write log data to DAT files
                RawRec(DATfiles.raw, &LogAng, &LogEpoch_t);
                ProcessRec(DATfiles.prc, &LogAng, &LogEpoch_t);

                //If new day close old & open new DAT file set
                time(&SysTime_t);
                if (localtime(&SysTime_t)->tm_mday != DATfiles.tm_mday)
                {
                    //Stop logging & close current data files
                    ReturnCode = RC_STOPFAIL;
                    if (stopLog(LogFile, &DATfiles) == FALSE) break;

                    //Initialize new data files & update day of month
                    ReturnCode = RC_DATCFAIL;
                    if (openDAT(&DATfiles, argv[SIZ_ARGV_BASE]) == NULL) break;

                    //Restart logging
                    ReturnCode = RC_LOGDFAIL;
                    if (startLog(LogFile) == FALSE) break;

                    //Fetch header & validate
                    ReturnCode = RC_LOGHFAIL;
                    if ((LogEpoch_t = readHeader(LogFile)) == (time_t) -1) break;

                    //Reset record counter & resume logging; Loop will wrap it back to 0
                    RecCount = (0 - 1);
                }
                else
                {
                    //When logging flush buffers for real-time access to records
                    fflush(DATfiles.raw);
                    fflush(DATfiles.prc);
                }

                //Sleep to prevent race condition
                sleep(WAIT_LOG_LOOP);
            }
        }
        printf("\n");

        //Determine if a footer record exists or if record read loop failed.
        if (feof(LogFile))
        {
            ReturnCode = RC_LOGNFTR;
            if (ftell(LogFile) != (long int) (sizeof(loghd) + (sizeof(logag) * RecCount)))\
                fseek(LogFile, (long int) (sizeof(logft) * -1), SEEK_END);
        }
        else break;

        //Read footer & validate
        if (fread((void *) &LogFtr, sizeof(logft), 1, LogFile))
        {
            ReturnCode = RC_LOGBFTR;
            if (LogFtr.rectype != RECTFTR) break;
            ReturnCode = RC_LOGRECC;
            if (LogFtr.numrecs != RecCount)
            {
                fprintf(DATfiles.log, FMT_ERR_FTRREC, RecCount, (unsigned int) LogFtr.numrecs);
                break;
            }
            ReturnCode = RC_SUCCESS;
        }

        break;
    }

    //Clean-up & exit
    if ((Cmd == CMD_LOG) && (LogFile)) stopLog(LogFile, &DATfiles);
    if (LogFile) fclose(LogFile);
    if (ReturnCode > RC_NOARG) fprintf(DATfiles.log, FMT_ERR_STDERR, prgName,\
                                       retCode[ReturnCode], errno);
    if (DATfiles.log != stderr) fclose(DATfiles.log); else printf(FMT_PRG_EXIT);

    return(ReturnCode);
} //End function main


/*Function getVolts.  Calculates the voltage represented by the adc value and scales.
 *
 *
 * Returns:     float value representing voltage.
 *
 * Parameters:  adc=value read by adc
 *              onev=adc value for one volt
 *              scale=integer representing the scale of the adc value
 *
 */
float getVolts(int adc, int onev, int scale)
{
    return(((float) ( (float) adc / (float) onev) * (float) scale));
} //End function getVolts


/*Function getAmps.  Calculates the amps represented by the adc value.
 *
 *
 * Returns:     float value representing amperes.
 *
 * Parameters:  adc=value read by adc
 *              onev=adc value for one volt
 *              offset=adc value representing 0A
 *              mVamp=float value representing sensor's mV per amp scale
 *
 */
float getAmps(int adc, int onev, int offset, float mVamp)
{
    return((getVolts(adc, onev, 1) - getVolts(offset, onev, 1)) / mVamp);
} //End function getAmps


/*Function TimeSinceEpoch.  Calculates the number of seconds since logger's epoch.
 *
 *
 * Returns:     integer representing the number of seconds since epoch 
 *              representing by object tse.
 *
 * Parameters:  tse=pointer to object of type logtm representing the epoch.
 *
 */
unsigned int TimeSinceEpoch(logtm *tse)
{
    return((tse->Min * 60) + tse->Sec);
} //End function TimeSinceEpoch


/*Function RawRec.  Formats the analog data in the object pointed to by record to
 * the FILE stream out.  No material alteration or calculations on the data are 
 * performed.
 *
 *
 * Returns:     Number of bytes written to out; a value of zero should be treated 
 *              as an error.
 *
 * Parameters:  out is a pointer to an object of type FILE representing the stream
 *              to write formated data to.
 *              record is a pointer to an object of type logag (analog record
 *              read from a logger file).
 *              epoch is a pointer to an object of time_t representing the epoch
 *              supplied by the header in the logger data file.
 *
 */
int RawRec(FILE *out, logag *record, time_t *epoch)
{
    static time_t calctm;
    static struct tm *formatm;
    int retval = 0;
    
    calctm = *epoch + (time_t) TimeSinceEpoch(&record->elapsed);
    formatm = localtime(&calctm);
    
    fprintf(out, FMT_OUT_RAWTM, record->seqnum, (formatm->tm_mon+1),\
            formatm->tm_mday, (formatm->tm_year + SYS_EPOCH_YEAR),\
            formatm->tm_hour, formatm->tm_min, formatm->tm_sec);
    
    if (record->pin == CH_2_CHARGE)\
        retval = fprintf(out, FMT_OUT_RAWVL, CHR_CHARGE,\
                         getAmps((int) record->adcval, (int) record->adcv,\
                                 (int) ADC_OFFSET, (float) ADC_MVAMP));
    else if (record->pin == CH_4_BATTERY)\
        retval = fprintf(out, FMT_OUT_RAWVL, CHR_BATTERY,\
                         getAmps((int) record->adcval, (int) record->adcv,\
                                 (int) ADC_OFFSET, (float) ADC_MVAMP));
    else if (record->pin == CH_5_VOLTS)\
        retval = fprintf(out, FMT_OUT_RAWVL, CHR_VOLTS,\
                         getVolts((int) record->adcval, (int) record->adcv,\
                                  ADC_SCALE_V));

    return(retval);
} //End function RawRec


/*Function getSample.  Aggregates data from individual logger records into 
 * discrete samples and augments with calculated fields.  Simple checks are done
 * to build a sample.  Records falling outside of limits are ignored.
 *
 * ADC data in each record is converted to voltage or power dependent on its
 * type.  When a full sample set is accumulated power calculations are added and
 * the sample is ready for consumption by the caller.
 *
 * Returns:     NULL is returned until a full sample is ready.
 *              A pointer to an array of adc objects is returned when the sample 
 *              is ready.
 *
 * Parameters:  result is a pointer to a caller allocated array of adc objects.
 *              The array must be at least ADC_SZ_SAMPLE in size.
 *              record is a pointer to an object of type logag (analog record
 *              ready from a logger file).
 *              epoch is a pointer to an object of time_t representing the epoch
 *              supplied by the header in the logger data file.
 *
 */
adc * getSample(adc *result, logag *record, time_t *epoch)
{
    static int addr = 0;
    static time_t calctm, lastm = 0;
    static adc sample[ADC_SZ_SAMPLE];
    static float amps_chg = 0, amps_bat = 0, volts = 0;
    
    //Calculate calendar date & time for the current record.
    calctm = *epoch + (time_t) TimeSinceEpoch(&record->elapsed);
    
    //Make sure variables are set on first call with new sample set.
    if (lastm == 0)
    {
        lastm = calctm;
        addr = 0;
        amps_chg = 0;
        amps_bat = 0;
        volts = 0;
        memset((void *) &sample, 0, (size_t) sizeof(sample));
    }
    
    if ((difftime(calctm, lastm) < (ADC_DIFF_MAX + 1)) && (addr < ADC_SZ_SAMPLE))
    {
        sample[addr].seqnum = record->seqnum;
        sample[addr].time = calctm;
        sample[addr].pinb = record->pin;
        if (record->pin == CH_2_CHARGE)
        {
            sample[addr].val_read = getAmps((int) record->adcval,\
                                            (int) record->adcv, (int) ADC_OFFSET,\
                                            (float) ADC_MVAMP);
            amps_chg = sample[addr].val_read;
            sample[addr].pinc = CHR_CHARGE;
        }
        else if (record->pin == CH_4_BATTERY)
        {
            sample[addr].val_read = getAmps((int) record->adcval,\
                                            (int) record->adcv, (int) ADC_OFFSET,\
                                            (float) ADC_MVAMP);
            amps_bat = sample[addr].val_read;
            sample[addr].pinc = CHR_BATTERY;
        }
        else if (record->pin == CH_5_VOLTS)
        {
            sample[addr].val_read = getVolts((int) record->adcval,\
                                             (int) record->adcv, ADC_SCALE_V);
            volts = sample[addr].val_read;
            sample[addr].pinc = CHR_VOLTS;
        }
        
        lastm = calctm;                         //Update for comparison on next call.
        ++addr;
        return(NULL);                           //Sample not ready.
    }
    
    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Only dump accumulated sample set if complete.  Maybe add code later to
     * log exceptions.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (addr == ADC_SZ_SAMPLE)
    {
        for (addr = 0; (addr < ADC_SZ_SAMPLE); ++addr)
        {
            //Calculated field for power from charger<->battery or battery->load
            if ((sample[addr].pinc == CHR_CHARGE) || (sample[addr].pinc == CHR_BATTERY))\
                sample[addr].val_calc = (sample[addr].val_read * volts);
            //Calculated field for NET power delivered to load
            else if (sample[addr].pinc == CHR_VOLTS)
            {
                sample[addr].val_calc = 0;
                if (amps_bat > 0)
                {
                    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                     * Charger > 0 must be added to net output since it's
                     * contributing.  If < 0 it means charger output is 0 (only
                     * drawing idle current to power unit) or it is discharging
                     * battery for maintenance.  Either way its draw is already
                     * accounted for in battery current.
                     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
                    sample[addr].val_calc = amps_bat;
                    if (amps_chg > 0) sample[addr].val_calc += amps_chg;
                }
                /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                 * If battery current < 0 then it is charging and charger is 
                 * carrying net load, if any.
                 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
                else if (amps_bat < 0) sample[addr].val_calc = amps_bat + amps_chg;
                sample[addr].val_calc *= sample[addr].val_read;
            }
        }
        //Copy data to user array
        memcpy((void *) result, (void *) &sample, sizeof(sample));
    }
    else result = NULL;                         //Bad sample
    
    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Dump of sample set complete.  To setup for next sample call function
     * recursively.  Necessary to prevent data loss since trigger was start of
     * new sample.  Clearing variable lastm ensures that new sample processing
     * begins.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    lastm = 0;
    getSample(result, record, epoch);
    
    return(result);
} //End function getSample


/*Function DumpRec.  Formats the analog data in the object pointed to by record
 * to the FILE stream out.  Records are organized into discrete samples with
 * power values added by getSample before they are output.
 *
 * Returns:     n/a
 *
 * Parameters:  out is a pointer to an object of type FILE representing the 
 *              stream to write formated data to.
 *              record is a pointer to an object of type logag (analog record 
 *              ready from a logger file).
 *              epoch is a pointer to an object of time_t representing the epoch
 *              supplied by the header in the logger data file.
 *
 */
void DumpRec(FILE *out, logag *record, time_t *epoch)
{
    static int addr = 0;
    static struct tm *formatm;
    static adc sample[ADC_SZ_SAMPLE];


    if (getSample((adc *) &sample, record, epoch))
    {
        for (addr = 0; (addr < ADC_SZ_SAMPLE); ++addr)
        {
            formatm = localtime(&sample[addr].time);
            fprintf(out, FMT_OUT_DUMP, sample[addr].seqnum, (formatm->tm_mon+1),\
                    formatm->tm_mday, (formatm->tm_year + SYS_EPOCH_YEAR),\
                    formatm->tm_hour, formatm->tm_min, formatm->tm_sec,\
                    sample[addr].pinc, sample[addr].val_read, sample[addr].val_calc);
        }
    }

    return;
} //End function DumpRec


/*Function ProcessRec.  Aggregates discrete samples into data sets of different
 * resolutions, averages data within a resolution, and outputs results.  Each
 * resolution represents a sample (time) interval.  Resolution A represents the
 * most granular, or shortest time interval.  B represents a longer interval and
 * represents the average resolution A data samples within its interval.
 *
 * Once data for a given resolution is complete a set of records are output.
 *
 * Returns:     n/a
 *
 * Parameters:  out is a pointer to an object of type FILE representing the 
 *              stream to write formated data to.
 *              record is a pointer to an object of type logag (analog record 
 *              ready from a logger file).
 *              epoch is a pointer to an object of time_t representing the epoch 
 *              supplied by the header in the logger data file.
 *
 */
void ProcessRec(FILE *out, logag *record, time_t *epoch)
{
    static int addr = 0;
    static struct tm *formatm;
    static adc sample[ADC_SZ_SAMPLE];

    static int cnt_res_a = 0, cnt_res_b = 0;
    static avg avg_res_a[ADC_SZ_SAMPLE], avg_res_b[ADC_SZ_SAMPLE];


    if (getSample((adc *) &sample, record, epoch))
    {
        for (addr = 0; (addr < ADC_SZ_SAMPLE); ++addr)
        {
            if (sample[addr].pinc == CHR_CHARGE)
            {
                avg_res_a[ADR_AVG_CHARGE].pinc = CHR_CHARGE;
                avg_res_b[ADR_AVG_CHARGE].pinc = CHR_CHARGE;
                avg_res_a[ADR_AVG_CHARGE].read += sample[addr].val_read;
                avg_res_a[ADR_AVG_CHARGE].calc += sample[addr].val_calc;
            }
            if (sample[addr].pinc == CHR_BATTERY)
            {
                avg_res_a[ADR_AVG_BATTERY].pinc = CHR_BATTERY;
                avg_res_b[ADR_AVG_BATTERY].pinc = CHR_BATTERY;
                avg_res_a[ADR_AVG_BATTERY].read += sample[addr].val_read;
                avg_res_a[ADR_AVG_BATTERY].calc += sample[addr].val_calc;
            }
            if (sample[addr].pinc == CHR_VOLTS)
            {
                avg_res_a[ADR_AVG_VOLTS].pinc = CHR_VOLTS;
                avg_res_b[ADR_AVG_VOLTS].pinc = CHR_VOLTS;
                avg_res_a[ADR_AVG_VOLTS].read += sample[addr].val_read;
                avg_res_a[ADR_AVG_VOLTS].calc += sample[addr].val_calc;
            }
        }

        if (++cnt_res_a == SPL_RES_A)
        {
            cnt_res_a = 0;
            formatm = localtime(&sample[ADC_SZ_SAMPLE - 1].time);
            for (addr = 0; (addr < ADC_SZ_SAMPLE); ++addr)
            {
                avg_res_a[addr].read /= (float) SPL_RES_A;
                avg_res_a[addr].calc /= (float) SPL_RES_A;
                avg_res_b[addr].read += avg_res_a[addr].read;
                avg_res_b[addr].calc += avg_res_a[addr].calc;
                fprintf(out, FMT_OUT_RESA, (formatm->tm_mon+1), formatm->tm_mday,\
                        (formatm->tm_year + SYS_EPOCH_YEAR), formatm->tm_hour,\
                        formatm->tm_min, formatm->tm_sec, avg_res_a[addr].pinc,\
                        avg_res_a[addr].read, avg_res_a[addr].calc);
            }
            memset((void *) &avg_res_a, 0, (size_t) sizeof(avg_res_a));
            
            if (++cnt_res_b == SPL_RES_B)
            {
                cnt_res_b = 0;
                for (addr = 0; (addr < ADC_SZ_SAMPLE); ++addr)
                {
                    avg_res_b[addr].read /= (float) SPL_RES_B;
                    avg_res_b[addr].calc /= (float) SPL_RES_B;
                    fprintf(out, FMT_OUT_RESB, (formatm->tm_mon+1), formatm->tm_mday,\
                            (formatm->tm_year + SYS_EPOCH_YEAR), formatm->tm_hour,\
                            formatm->tm_min, formatm->tm_sec, avg_res_b[addr].pinc,\
                            avg_res_b[addr].read, avg_res_b[addr].calc);
                }
                memset((void *) &avg_res_b, 0, (size_t) sizeof(avg_res_b));
            }
        }
    }

    return;
} //End function ProcessRec


/*Function openDAT.  Creates raw & processed output data & log files for writing.
 * The files are located in the path pointed to by base and named according to base
 * with the current date appended.  A complete file path is 
 * <file>-T-YYYYMMDD.EEE where:
 *      T=R for raw or P for processed
 *      YYYY=current year
 *      MM=current month
 *      DD=current day
 *      EEE=DAT for data, LOG for log file
 *
 * A log file is only created if the log pointer in object datstream is stderr.
 * This way a single log file is used for entire session.
 *
 * Returns:     pointer to datstream if successful or NULL on error; errno should
 * be examined before & after call for details since this function may fail for
 * reasons other than file operations.  In that case it should be assumed that 
 * problems exist with the user-supplied base.
 *
 * Parameters:  base=char * to the user-supplied path & file name base; it must 
 *              be NULL terminated.
 *              datstream=pointer to object of type dats to store the FILE stream
 *              pointers in.  On error the contents are indeterminate.
 *
 */
dats *openDAT(dats *datstream, char *base)
{
    static char *dpfmt = FMT_DAT_PATH, *extdat = STR_EXT_DAT,\
    *extlog = STR_EXT_LOG, *dfmode = FMODE_DAT;

    dats *retval = NULL;
    char filename[SIZ_LOG_PATH];
    time_t systime;
    struct tm *ltime;

    //Clear stream pointers
    datstream->raw = NULL;
    datstream->prc = NULL;

    //Waterfall processing loop
    while (TRUE)
    {
        //Limit check on length of base.
        if (strnlen(base, (size_t) SIZ_MAX_BASE) > (size_t) (SIZ_MAX_BASE -2)) break;

        //Fetch local time; store day of month in object pointed to by datstream
        time(&systime);
        ltime = localtime(&systime);
        datstream->tm_mday = ltime->tm_mday;

        //Only open a log file if stream is stderr
        if (datstream->log == stderr)
        {
            snprintf(filename, (size_t) SIZ_LOG_PATH, dpfmt, base, CHR_LOG_BEGIN,\
                     (ltime->tm_year + SYS_EPOCH_YEAR), (ltime->tm_mon + 1),\
                     ltime->tm_mday, extlog);
            if ((datstream->log = fopen(filename, dfmode)) == NULL)
            {
                datstream->log = stderr;
                break;
            }
        }

        //Open raw data output file & print field header
        snprintf(filename, (size_t) SIZ_LOG_PATH, dpfmt, base, CHR_LOG_RAW,\
                 (ltime->tm_year + SYS_EPOCH_YEAR), (ltime->tm_mon + 1),\
                 ltime->tm_mday, extdat);
        if ((datstream->raw = fopen(filename, dfmode)) == NULL) break;
        fprintf(datstream->raw, STR_HDR_RAW);

        //Open process data output file & print field header
        snprintf(filename, (size_t) SIZ_LOG_PATH, dpfmt, base, CHR_LOG_PROC,\
                 (ltime->tm_year + SYS_EPOCH_YEAR), (ltime->tm_mon + 1),\
                 ltime->tm_mday, extdat);
        if ((datstream->prc = fopen(filename, dfmode)) == NULL)\
            fclose(datstream->raw); else retval = datstream;
        fprintf(datstream->prc, FMT_HDR_FIELDS, STR_HDR_PRC, STR_HDR_DMP);

        break;
    }

    return(retval);
} //End function openDAT


/*Function startLog.  Attempts to establish a known working state with logger and
 * initiate logging.  This is problematic mainly because there is no facility to
 * determine if data is waiting to be read from the stream.  Since the stream 
 * read functions are blocking they don't return until a character is ready.
 *
 * This function worsk around this by sending a series of commands that attempt 
 * to ensure that the logger is in command mode and binary output is on.  The ID
 * command is issued so that the last expected response is the processor (PIC MPU)
 * ID.  The output stream is read seaching for this to avoid reading past the end
 * of available data (to avoid a hung state) and confirm that logger is waiting 
 * for a command.  Once confirmed the start logging command is issued.
 *
 *
 * Returns:     TRUE if start logging was successful
 *              FALSE if logging could not be started; logger state unknown
 *
 * Parameters:  out is a pointer to an object of type FILE representing the device
 *              stream to communicate with logger
 *
 */
int startLog(FILE *out)
{
    time_t systime;
    struct tm *ltime;
    unsigned char idstr[SIZ_BUF_IDS];
    int cmdpos, idpos, retval;

    for (retval = FALSE; (TRUE);)                           //Waterfall loop
    {
        if (snprintf((char *) idstr, (size_t) SIZ_BUF_IDS, FMT_BUF_IDS,\
                     (unsigned int) PROC_ID) < 0) break;

        if (fprintf(out, FMT_LOG_HDSK) < 0) break;
        for (cmdpos=SIZ_RSP_CMD, idpos = 0; (cmdpos); --cmdpos, ++idpos)
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Look for processor ID; this marks end of logger output in command 
             * mode and puts it in known good state to issue 'start logging' 
             * command.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if ((char) fgetc(out) != idstr[idpos]) idpos = 0;
            if (idpos == (SIZ_PROC_ID - 1)) break;
        }
        if (!(cmdpos)) break;       //Didn't find ID - unknown response / state!

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Fetch system time, update logger epoch, and start logging.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        time(&systime);
        ltime = localtime(&systime);
        if (fprintf(out, FMT_LOG_STRT,\
                    (ltime->tm_year - ((ltime->tm_year / 100) * 100)),\
                    (ltime->tm_mon + 1), ltime->tm_mday, ltime->tm_hour,\
                    ltime->tm_min, ltime->tm_sec) < 0) break;

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Consume command response up to start of header record.  Push first
         * byte back onto stream.  If not found in size of header record bytes
         * then problem.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        for (cmdpos = SIZ_RSP_CMD; (cmdpos); --cmdpos)\
            if (fgetc(out) == (int) RECTHDR) break;
        if (!(cmdpos)) break;
        ungetc((int) RECTHDR, out);

        retval = TRUE;
        break;
    }

    return(retval);
} //End function startLog


/*Function stopLog.  Attempts to stop logging and confirm logger is in command
 * mode.  This is problematic mainly because there is no facility to determine
 * if data is waiting to be read from the stream.  Since the stream read functions
 * are blocking they don't return until a character is ready.
 *
 * After the stop logging command is sent, the ID command is issued so that the
 * last expected response is the processor (PIC MPU) ID.  The output stream is
 * read seaching for this to avoid reading past the end of available data (to
 * avoid a hung state) and confirm that logger is waiting for a command.
 *
 * This function also closes the current DAT output files and clears the stream 
 * handles.
 *
 *
 * Returns:     TRUE if successful
 *              FALSE if logging could not be stopped; logger state unknown
 *
 * Parameters:  out=pointer to an object of type FILE representing the device
 *              stream to communicate with logger
 *              datstream=pointer to object of type dats containing DAT FILE 
 *              stream pointers
 *
 */
int stopLog(FILE *out, dats *datstream)
{
    unsigned char idstr[SIZ_BUF_IDS];
    int cmdpos, idpos, retval;
    
    //First attempt to close DAT files & clear stream pointers
    if (datstream->raw) fclose(datstream->raw);
    if (datstream->prc) fclose(datstream->prc);
    datstream->raw = NULL;
    datstream->prc = NULL;

    //Now tell logger to stop
    for (retval = FALSE; (TRUE);)                           //Waterfall loop
    {
        if (snprintf((char *) idstr, (size_t) SIZ_BUF_IDS, FMT_BUF_IDS, (unsigned int) PROC_ID) < 0) break;
        
        if (fprintf(out, FMT_LOG_STOP) < 0) break;

        //Wait for logging to stop before issuing ID command
        sleep(WAIT_LOG_STOP);
        if (fprintf(out, FMT_LOG_GID) < 0) break;

        //Now parse responses
        for (cmdpos=SIZ_RSP_CMD, idpos = 0; (cmdpos); --cmdpos, ++idpos)
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Look for processor ID; this marks end of logger output & confirms 
             * it's in command mode.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if ((char) fgetc(out) != idstr[idpos]) idpos = 0;
            if (idpos == (SIZ_PROC_ID - 1)) break;
        }
        if (!(cmdpos)) break;                               //This means ID string wasn't found - unknown response / state!

        retval = TRUE;
        break;
    }
    
    return(retval);
} //End function stopLog


/*Function sigHandler.  Signal handler for user interrupt & TERM signals.  Attempts to ensure that open files / devices are closed.  If in
 * 'log' mode logger is placed in command mode.
 *
 * Returns:     n/a
 *
 * Parameters:  signal=signal #; refer to signal.h.
 *
 */
void sigHandler(int signal)
{
    static int recurse = 1;

    if (recurse)
    {
        recurse = 0;
        stopLog(LogFile, &DATfiles);
        fclose(LogFile);

        fprintf(DATfiles.log, FMT_TRM_SIGNAL, prgName, signal);
        if (DATfiles.log != stderr) fclose(DATfiles.log);   //Close after logging sig term message
        exit(RC_SIGEXIT);
    }

    return;
} //End function sigHandler


/*Function readHeader.  Reads logger header from stream logfile, validates, and
 * converts logger's epoch to calendar time.
 *
 * Returns:     On success a time_t value representing logger epoch in calendar 
 *              time.  Otherwise -1 which could indicate a file operation
 *              error, bad header data, or failure to convert the epoch to a
 *              time_t value.  errno should be checked before & after the
 *              function call for detailed error handling.
 *
 * Parameters:  logfile=pointer to object of type FILE representing the stream
 *              that the header will be read from.
 *
 */
time_t readHeader(FILE *logfile)
{
    time_t SysTime_t;                               //Used to fetch system time
    struct tm LogEpoch;                             //Time that logger was initialized
    loghd LogHdr;                                   //Header record
    time_t retval = -1;

    while (TRUE)
    {
        //Read header record & validate
        if (fread((void *) &LogHdr, sizeof(loghd), SIZ_REC_HDR, logfile) != SIZ_REC_HDR) break;
        if (LogHdr.rectype != RECTHDR) break;
        if (LogHdr.deviceid.word != PROC_ID) break;

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Fetch logger epoch, store, and get tm value as local time.  First get
         * current system time so that DST flag can be properly set.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        time(&SysTime_t);
        LogEpoch.tm_isdst = localtime(&SysTime_t)->tm_isdst;
        LogEpoch.tm_sec = (int) LogHdr.epoch.second;
        LogEpoch.tm_min = (int) LogHdr.epoch.minute;
        LogEpoch.tm_hour = (int) LogHdr.epoch.hour;
        LogEpoch.tm_year = (int) ((LogHdr.epoch.year + LOG_EPOCH_YEAR) - SYS_EPOCH_YEAR);
        LogEpoch.tm_mon = (int) (LogHdr.epoch.month - 1);
        LogEpoch.tm_mday = (int) LogHdr.epoch.day;

        retval = mktime(&LogEpoch);
        break;
    }
    
    return(retval);
} //End function readHeader



//End file.